#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/times.h>


#include "mci.h"

#define EMPTY 0
#define OCCUPIED_A 1
#define OCCUPIED_B -1
#define MOL_A    1
#define MOL_B    -1



short int pos_x[PART_ANZ],pos_y[PART_ANZ],pos_z[PART_ANZ],typ[PART_ANZ];
short int space[SYS_SIZE][SYS_SIZE][SYS_SIZE];




void init_rnd(void);
int   rnd(void);
double rndf(void);
void init_parts(void);
 void sim(int trun);
int total_energy(void);
int add(int x,int d);

FILE *fpl,*fpt;
unsigned int seed,os;


int main(int argc, char *argv[])
{
 double runTime,runSecs,runMin;
 int trun=-1;
 char file[256];
 struct tms start,stop;


 trun = atoi(argv[1]);

 sprintf(file,"result.%s",argv[2]);

 if(trun<=0)
  {
   printf("MC-Bench needs TIME_STEPS to be specified in mcb.conf\n");
   printf("Terminating\n");
   return(-1);
  }
  
 printf("running MCIBench with %d particles for %d steps\n",PART_ANZ,trun);

 init_rnd();
 seed =1234567l;
 fpl = fopen("mci.log","w");
 if(fpl == NULL)
   printf("Can't open outputfile\n");
 init_parts();

#ifdef WITH_PRINTF
 printf("Starting  MCI-Bench\n");
#endif
 times(&start);
 sim(trun);
 times(&stop);

 runTime = ((stop.tms_stime +stop.tms_utime) - (start.tms_stime +start.tms_utime)) / (double)sysconf(_SC_CLK_TCK);
 runSecs = fmod(runTime ,60.);
 runMin  = runTime-runSecs;
 fpt=fopen(file,"w");
 printf("Laufzeit %f:%f\n",runMin,runSecs);
 fprintf(fpt,"%d %f %d\n",trun,runTime,PART_ANZ);
 fclose(fpl);
 fclose(fpt);

}

/***************************************************/
/* Initialisierung der Zufallsgenerator Wortl"ange */
/***************************************************/
void init_rnd() 
{ 
 os   = (unsigned long) 256*256*256*128; 
} 

/***************************************/
/* Erzeugung von integer Zufallszahlen */
/***************************************/
int rnd() 
{ 
 unsigned int j; 
 
 j     = (69069l* seed+1) ; 
 seed = j; 
 return(j); 
} 

/***************************************/
/* Erzeugung von double  Zufallszahlen */
/***************************************/
double rndf() 
{ 
 unsigned int j; 
 double r; 
  
 j     = (69069l* seed+1); 
 seed = j; 
 r     = (double)j/os*.5; 
 return(r); 
} 

void init_parts(void)
{
 short x,y,z;
 int i,j,k;

 for(i=0;i<SYS_SIZE; i++)
  for(j=0;j<SYS_SIZE; j++)
   for(k=0;k<SYS_SIZE; k++)
     space[i][j][k]=EMPTY;

 for(i=0;i<PART_ANZ; i++)
  {
   x = (int) (rndf()*SYS_SIZE);
   y = (int) (rndf()*SYS_SIZE);
   z = (int) (rndf()*SYS_SIZE);
   while(space[x][y][z]!=EMPTY)
    {
     x = (int) (rndf()*SYS_SIZE);
     y = (int) (rndf()*SYS_SIZE);
     z = (int) (rndf()*SYS_SIZE);
    }
   /*   printf("part %d : %4d%4d%4d \t\t%d\n",i,x,y,z,space[10][0][4]);*/
   pos_x[i]=x;
   pos_y[i]=y;
   pos_z[i]=z;
   if(rndf()>=0.5)
    {
     typ[i] =MOL_A;
     space[x][y][z]=OCCUPIED_A;

    }
   else
    {
     typ[i]=MOL_B;
     space[x][y][z]=OCCUPIED_B;
    }
  }
}

void sim(int trun)
{
 int t,i;
 int x,y,z;
 int xa,ya,za;
 int dx,dy,dz;
 int en,ea,de,e,dn;
 
 for(t=0; t<trun; t++)
  {
   for(i=0; i<PART_ANZ; i++)
    {
     dx=(int)(3*rndf()) -1;
     dy=(int)(3*rndf()) -1;
     dz=(int)(3*rndf()) -1;
     while(dx==dy==dz==0)
      {
       dx=(int)(3*rndf()) -1;
       dy=(int)(3*rndf()) -1;
       dz=(int)(3*rndf()) -1;
      }
     
     x = add(pos_x[i],dx);
     y = add(pos_y[i],dy);
     z = add(pos_z[i],dz);
     if(space[x][y][z] == EMPTY)
      {
       xa=pos_x[i];
       ya=pos_y[i];
       za=pos_z[i];
       ea=en=0;
       ea = typ[i]*space[add(xa,1)][ya][za] + typ[i]*space[add(xa,-1)][ya][za] +
            typ[i]*space[xa][add(ya,1)][za] + typ[i]*space[xa][add(ya,-1)][za] +
            typ[i]*space[xa][ya][add(za,1)] + typ[i]*space[xa][ya][add(za,-1)];
       en = typ[i]*space[add(x,1)][y][z] + typ[i]*space[add(x,-1)][y][z] +
            typ[i]*space[x][add(y,1)][z] + typ[i]*space[x][add(y,-1)][z] +
            typ[i]*space[x][y][add(z,1)] + typ[i]*space[x][y][add(z,-1)];
       dn = en -ea;
       if(dn<0)
	    {
	     space[xa][ya][za]=EMPTY;
	     pos_x[i]=x; pos_y[i]=y; pos_z[i]=z;
	     space[x][y][z]=typ[i];
   	    }
       else
	    {
	     if(rndf() < exp(-((double)dn/AKT_TEMP)) )
 	      {
	       space[xa][ya][za]=EMPTY;
	       pos_x[i]=x; pos_y[i]=y; pos_z[i]=z;
	       space[x][y][z]=typ[i];
	      }
	    }
      }
    }
   if( ((t+1) % TLOG) == 0)
    {
     e = total_energy();
     fprintf(fpl,"%6d) E/kT=%12g \n",t+1,(double)e*AKT_TEMP);
#ifdef WITH_PRINTF
     printf("%6d) E/kT=%12g \n",t+1,(double)e*AKT_TEMP);
#endif
    }
  }
}
int total_energy(void)
{
 int e=0,i,de;
 int x,y,z;

 for(i=0; i<PART_ANZ; i++)
  {
   x=pos_x[i];
   y=pos_y[i];
   z=pos_z[i];
   de = typ[i]*space[add(x,-1)][y][z] +
    typ[i]*space[x][add(y,-1)][z] +
    typ[i]*space[x][y][add(z,-1)] ;  
   e+=de;
  }
 return(e);
}

int add(int x,int d)
{
 if((x+d)>=SYS_SIZE)
  return(x+d- SYS_SIZE);
 if ((x+d)<0)
  return(x+d+ SYS_SIZE);
 return(x+d);
}

